import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.WindowObject;
import std.stdio;
import std.string;
import std.utf;
import std.math;

import SliderBase ;
import SliderGrad ;
import Spin;

COLORREF RGBA(uint r, uint g, uint b, uint a) {
  return cast(COLORREF)(r | (g << 8) | (b << 16) | (a<<24));
}

enum CMD{
  SLID_R = 0x100,
  SLID_G = 0x101,
  SLID_B = 0x102,
  SLID_A = 0x103,
  SPIN_R = 0x110,
  SPIN_G = 0x111,
  SPIN_B = 0x112,
  SPIN_A = 0x113
} ;

class RGBslider : WindowObject {
  SliderGrad  slider_r ;
  SliderGrad  slider_g ;
  SliderGrad  slider_b ;
  SliderGrad  slider_a ;
  Spin        spin_r   ;
  Spin        spin_g   ;
  Spin        spin_b   ;
  Spin        spin_a   ;
  bool        enbAlpha=false ;
  uint        bg_color  = RGBA(192,192,192,255) ;
  HFONT       hfont;

  this(string name, string ClassName) {
    winc.style          = 0;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.lpszMenuName   = null;
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    winc.hbrBackground  = CreateSolidBrush(GetSysColor(COLOR_BTNFACE)) ;
    winc.lpszClassName  = toUTFz!(wchar*)(ClassName) ;
    super(name) ;
    this.registwinc() ;
  }

  void setBGCOLOR(int r, int g, int b){
    bg_color = RGBA(r,g,b,255) ;
  }

  void enableAlpha(bool enbAlpha){
    this.enbAlpha=enbAlpha ;
  }

  override LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      DeleteObject(hfont);
      break ;

    case WM_CREATE:
      hfont = CreateFont(
                         -12 , 0 , 0 , 0 , FW_REGULAR , FALSE , FALSE , FALSE ,
                         DEFAULT_CHARSET, OUT_DEFAULT_PRECIS ,
                         CLIP_DEFAULT_PRECIS , DEFAULT_QUALITY ,
                         FIXED_PITCH, null
                         );

      spin_r = new Spin("SpinRed", "SpinRed") ;
      spin_r.setStyle( WS_CHILD |WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE ) ;
      spin_r.setPosition(0, 4,40,20) ;
      spin_r.setTextFormat("%.0f") ;
      spin_r.setInfo( 0, 0, 255, 1 ) ;
      addObject(spin_r, cast(HMENU)(CMD.SPIN_R), false, false) ;
      spin_r.setTextFont(hfont) ;

      spin_g = new Spin("SpinGreen", "SpinGreen") ;
      spin_g.setStyle( WS_CHILD |WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE ) ;
      spin_g.setPosition(0,24,40,20) ;
      spin_g.setTextFormat("%.0f") ;
      spin_g.setInfo( 0, 0, 255, 1 ) ;
      addObject(spin_g, cast(HMENU)(CMD.SPIN_G), false, false) ;
      spin_g.setTextFont(hfont) ;

      spin_b = new Spin("SpinBlue", "SpinBlue") ;
      spin_b.setStyle( WS_CHILD |WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE ) ;
      spin_b.setPosition(0,44,40,20) ;
      spin_b.setTextFormat("%.0f") ;
      spin_b.setInfo( 0, 0, 255, 1 ) ;
      addObject(spin_b, cast(HMENU)(CMD.SPIN_B), false, false) ;
      spin_b.setTextFont(hfont) ;

      if( enbAlpha==true ){
        spin_a = new Spin("SpinAlpha", "SpinAlpha") ;
        spin_a.setStyle( WS_CHILD |WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE ) ;
        spin_a.setPosition(0,64,40,20) ;
        spin_a.setTextFormat("%.0f") ;
        spin_a.setInfo( 0, 0, 255, 1 ) ;
        addObject(spin_a, cast(HMENU)(CMD.SPIN_A), false, false) ;
        spin_a.setTextFont(hfont) ;
      }

      slider_r = new SliderGrad("SliderRed","SliderRed") ;
      slider_r.addExStyle(WS_EX_CLIENTEDGE) ;
      slider_r.setStyle(WS_CHILD |WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE) ;
      slider_r.setPosition(40,6,100,16) ;
      slider_r.setInfo(0, 0.0, 0.0, 255.0) ;
      slider_r.setBGcolor(  0,  0,  0,255) ;
      slider_r.setFGcolor(255,  0,  0,255) ;
      addObject(slider_r, cast(HMENU)(CMD.SLID_R),false,false) ;

      slider_g = new SliderGrad("SliderGreen","SliderGreen") ;
      slider_g.addExStyle(WS_EX_CLIENTEDGE) ;
      slider_g.setStyle(WS_CHILD |WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE) ;
      slider_g.setPosition(40,26,100,16) ;
      slider_g.setInfo(0, 0.0, 0.0, 255.0) ;
      slider_g.setBGcolor(  0,  0,  0,255) ;
      slider_g.setFGcolor(  0,255,  0,255) ;
      addObject(slider_g, cast(HMENU)(CMD.SLID_G),false,false) ;

      slider_b = new SliderGrad("SliderBlue","SliderBlue") ;
      slider_b.addExStyle(WS_EX_CLIENTEDGE) ;
      slider_b.setStyle(WS_CHILD |WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE) ;
      slider_b.setPosition(40,46,100,16) ;
      slider_b.setInfo(0, 0.0, 0.0, 255.0) ;
      slider_b.setBGcolor(  0,  0,  0,255) ;
      slider_b.setFGcolor(  0,  0,255,255) ;
      addObject(slider_b, cast(HMENU)(CMD.SLID_B),false,false) ;

      if( enbAlpha==true ){
        slider_a = new SliderGrad("SliderAlpha","SliderAlpha") ;
        slider_a.addExStyle(WS_EX_CLIENTEDGE) ;
        slider_a.setStyle(WS_CHILD |WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE) ;
        slider_a.setPosition(40,66,100,16) ;
        slider_a.setInfo(0, 0.0, 0.0, 255.0) ;
        slider_a.setBGcolor(255,255,255,  0) ;
        slider_a.setFGcolor(255,255,255,255) ;
        addObject(slider_a, cast(HMENU)(CMD.SLID_A),false,false) ;
      }

      break ;

    case WM_PAINT:
      HDC  hdc;
      PAINTSTRUCT ps ;
      hdc   = BeginPaint(hwnd, &ps);
      EndPaint(hwnd, &ps);
      break ;

    case WM_MOVE:
      RedrawWindow( hwnd, null, null, RDW_INVALIDATE) ;
      break ;

    case WM_SIZE:
      MoveWindow(slider_r.getHWND(),40, 6,winW-40,16,true) ;
      MoveWindow(slider_g.getHWND(),40,26,winW-40,16,true) ;
      MoveWindow(slider_b.getHWND(),40,46,winW-40,16,true) ;
      if( enbAlpha==true ){
        MoveWindow(slider_a.getHWND(),40,66,winW-40,16,true) ;
      }
      RedrawWindow( hwnd, null, null, RDW_INVALIDATE) ;
      break;

    case WM_COMMAND:
      switch( LOWORD(wp) ){
      case CMD.SPIN_R :
        slider_r.setValue( spin_r.getValue() ) ;
        syncAlphaSlider() ;
        sendmsgtoparent(WM_COMMAND, cast(int)(getHMENU()), 0) ;
        break;
      case CMD.SPIN_G :
        slider_g.setValue( spin_g.getValue() ) ;
        syncAlphaSlider() ;
        sendmsgtoparent(WM_COMMAND, cast(int)(getHMENU()), 0) ;
        break;
      case CMD.SPIN_B :
        slider_b.setValue( spin_b.getValue() ) ;
        syncAlphaSlider() ;
        sendmsgtoparent(WM_COMMAND, cast(int)(getHMENU()), 0) ;
        break;
      case CMD.SPIN_A :
        slider_a.setValue( spin_a.getValue() ) ;
        sendmsgtoparent(WM_COMMAND, cast(int)(getHMENU()), 0) ;
        break;
      case CMD.SLID_R :
        spin_r.setValue( round(slider_r.getValue()) ) ;
        syncAlphaSlider() ;
        sendmsgtoparent(WM_COMMAND, cast(int)(getHMENU()), 0) ;
        break;
      case CMD.SLID_G :
        spin_g.setValue( round(slider_g.getValue()) ) ;
        syncAlphaSlider() ;
        sendmsgtoparent(WM_COMMAND, cast(int)(getHMENU()), 0) ;
        break;
      case CMD.SLID_B :
        spin_b.setValue( round(slider_b.getValue()) ) ;
        syncAlphaSlider() ;
        sendmsgtoparent(WM_COMMAND, cast(int)(getHMENU()), 0) ;
        break;
      case CMD.SLID_A :
        spin_a.setValue( round(slider_a.getValue()) ) ;
        sendmsgtoparent(WM_COMMAND, cast(int)(getHMENU()), 0) ;
        break;
      default :        break ;
      }
      break ;

    default :
      return( DefWindowProc(hwnd , msg , wp , lp) );
    }
    return(0);
  }

  private int syncAlphaSlider(){
    if( enbAlpha==true ){
      uint col=getRGB() ;
      int r = (col    )&0x000000ff ;
      int g = (col>>8 )&0x000000ff ;
      int b = (col>>16)&0x000000ff ;
      slider_a.setBGcolor(r,g,b,  0) ;
      slider_a.setFGcolor(r,g,b,255) ;
      slider_a.fillGrad() ;
      MoveWindow(slider_a.getHWND(),40,66,winW-40,16,true) ;
      RedrawWindow(slider_a.getHWND(), null, null, RDW_INVALIDATE|RDW_UPDATENOW) ;
    }
    return(0) ;
  }

  int setRGB(uint col){
    if( getHWND() !is null ){
      int r = (col    )&0x000000ff ;
      int g = (col>>8 )&0x000000ff ;
      int b = (col>>16)&0x000000ff ;
      slider_r.setValue(r) ; spin_r.setValue(r) ;
      slider_g.setValue(g) ; spin_g.setValue(g) ;
      slider_b.setValue(b) ; spin_b.setValue(b) ;
      syncAlphaSlider() ;
      RedrawWindow( getHWND(), null, null, RDW_INVALIDATE|RDW_UPDATENOW) ;
    }
    return(0) ;
  }

  uint getRGB(){
    uint col = (cast(uint)(spin_r.getValue())&0x000000ff)    |
               (cast(uint)(spin_g.getValue())&0x000000ff)<<8 |
               (cast(uint)(spin_b.getValue())&0x000000ff)<<16;

    return(col) ;
  }

  int setRGBA(uint col){
    if( getHWND() !is null ){
      int r = (col    )&0x000000ff ;
      int g = (col>>8 )&0x000000ff ;
      int b = (col>>16)&0x000000ff ;
      int a = (col>>24)&0x000000ff ;
      slider_r.setValue(r) ; spin_r.setValue(r) ;
      slider_g.setValue(g) ; spin_g.setValue(g) ;
      slider_b.setValue(b) ; spin_b.setValue(b) ;
      if( enbAlpha==true ){
        slider_a.setValue(a) ; spin_a.setValue(a) ;
      }
      syncAlphaSlider() ;
      RedrawWindow( getHWND(), null, null, RDW_INVALIDATE|RDW_UPDATENOW) ;
    }
    return(0) ;
  }

  uint getRGBA(){
    uint alpha=0xff000000 ;
    if( enbAlpha==true ){
      alpha=(cast(uint)(spin_a.getValue())&0x000000ff)<<24 ;
    }
    uint col = alpha |
               (cast(uint)(spin_r.getValue())&0x000000ff)    |
               (cast(uint)(spin_g.getValue())&0x000000ff)<<8 |
               (cast(uint)(spin_b.getValue())&0x000000ff)<<16;

    return(col) ;
  }
}
