import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.DialogBoxClass;
import win32frame.CheckBox;
import win32frame.PushButton;
import std.stdio;
import std.string;
import std.utf;

import FontMan;
import Spin;

enum CMD{
  OK     = 0x100,
  CANCEL = 0x101,
  CHECK  = 0x201,
  SPIN   = 0x301,
} ;

class TestDialog : DialogBoxClass {
  PushButton  btnok  ;
  PushButton  btncan ;
  CheckBox    chkbox ;
  Spin        spin   ;
  HFONT       hfont_f12;
  HICON       hicon  ;

  int         dlgstate ;
  int         chkstate ;
  double      spinvalue;
  const int   defaultWidth =204 ;
  const int   defaultHeight=200 ;
  const DWORD defaultStyle =WS_SIZEBOX | WS_VISIBLE | WS_SYSMENU ;


  this(string name, string ClassName) {
    UnregisterClass( toUTFz!(wchar*)(ClassName), GetModuleHandle(null) ) ;
    winc.cbSize         = WNDCLASSEX.sizeof;
    winc.style          = 0;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.hIcon          = null ;
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    winc.hbrBackground  = CreateSolidBrush(GetSysColor(COLOR_BTNFACE)) ;
    winc.lpszMenuName   = null;
    winc.lpszClassName  = toUTFz!(wchar*)(ClassName) ;
    super(name) ;
  }

  int Create( int x, int y, HWND phwnd, HMENU hmenu){
    return( create(defaultStyle,x,y,defaultWidth,defaultHeight,phwnd,hmenu) ) ;
  }

  override LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      DeleteObject(hfont_f12);
      DeleteObject(hicon) ;
      break ;
    case WM_CREATE:
      FontMan font_f12 ; font_f12.init(-12,FIXED_PITCH) ;

      hfont_f12 = font_f12.createHFONT() ;

      hicon = cast(HICON)(LoadImage(GetModuleHandle(null), toUTFz!(wchar*)("WAPICON"),
                                    IMAGE_ICON,
                                    GetSystemMetrics(SM_CXICON),
                                    GetSystemMetrics(SM_CYICON),
                                    LR_DEFAULTCOLOR)) ;

      btnok= new PushButton("OK") ;
      btnok.setPosition(0,0,100,20) ;
      btnok.addStyle(WS_TABSTOP|BS_DEFPUSHBUTTON) ;
      addObject( btnok, cast(HMENU)(CMD.OK) ) ;
      btnok.setFont(hfont_f12) ;

      btncan= new PushButton("Cancel") ;
      btncan.setPosition(0,20,100,20) ;
      btncan.addStyle(WS_TABSTOP) ;
      addObject( btncan, cast(HMENU)(CMD.CANCEL) ) ;
      btncan.setFont(hfont_f12) ;

      chkbox  = new CheckBox("Check") ;
      chkbox.setPosition(0,40,100,20) ;
      chkbox.addStyle(BS_AUTOCHECKBOX) ;
      addObject( chkbox, cast(HMENU)(CMD.CHECK) ) ;
      chkbox.setFont(hfont_f12) ;
      chkstate=0 ;

      spin = new Spin("Spin", "Spin") ;
      spin.setStyle( WS_CHILD |WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE ) ;
      spin.addStyle(WS_TABSTOP) ;
      spin.setPosition(100,40,90,20) ;
      spin.setTextFormat("%.3f") ;
      spinvalue=0.0 ;
      spin.setInfo( spinvalue, 0, 100, 1 ) ;
      addObject( spin, cast(HMENU)(CMD.SPIN) ) ;
      spin.setTextFont(hfont_f12) ;

      SetFocus(btnok.getHWND()) ;

      dlgstate=CMD.CANCEL ;
      break ;

    case WM_PAINT:
      PAINTSTRUCT ps;
      HDC hdc;
      hdc = BeginPaint(hwnd, &ps);
      SetBkMode(hdc,TRANSPARENT) ;
      SelectObject(hdc , hfont_f12);
      TextOut(hdc , 8 , 60 , toUTFz!(wchar*)("ダイアログテスト") , 8);
      DrawIcon(hdc, 80,80, hicon) ;
      EndPaint(hwnd, &ps);
      break ;

    case WM_SIZE:
      break ;

    case WM_CLOSE:
      setResult(dlgstate) ;
      DestroyWindow(hwnd);
      break ;

    case WM_COMMAND:
      dlgstate=LOWORD(wp) ;
      switch( dlgstate ){
      case CMD.OK :
      case CMD.CANCEL :
        setResult(dlgstate) ;
        DestroyWindow(hwnd);
        break ;
      case CMD.CHECK :
        chkstate=chkbox.getState() ;
        break ;
      case CMD.SPIN :
        spinvalue=spin.getValue() ;
        break ;
      default :
        break ;
      }
      break ;

    default :
      return( DefDlgProc(hwnd , msg , wp , lp) );
    }
    return(0);
  }

  int getCheckState(){
    return(chkstate) ;
  }

  double getSpinValue(){
    return(spinvalue) ;
  }
}
