import std.stdio;
import std.string;
import std.conv;
import std.file;

struct BitmapFileHeader
{
  ushort bfType ;
  uint   bfSize ;
  ushort bfReserved1 ;
  ushort bfReserved2 ;
  uint   bfOffBits ;
}

struct CIEXYZ
{
  uint ciexyzX ;
  uint ciexyzY ;
  uint ciexyzZ ;
}

struct CIEXYZTRIPLE
{
    CIEXYZ ciexyzRed;
    CIEXYZ ciexyzGreen;
    CIEXYZ ciexyzBlue;
}

enum {
  BI_RGB       =0 ,
  BI_RLE8      =1 ,
  BI_RLE4      =2 ,
  BI_BITFIELDS =3 ,
}

struct BitmapInformationHeader
{
  uint   Size ;
  int    Width ;
  int    Height ;
  ushort Planes ;
  ushort BitCount ;
  uint   Compression ;
  uint   SizeImage ;
  int    XPixPerMeter ;
  int    YPixPerMeter ;
  uint   ClrUsed ;
  uint   CirImportant ;

  uint   RedMask;
  uint   GreenMask;
  uint   BlueMask;
  uint   AlphaMask;
  uint   CSType;
  CIEXYZTRIPLE Endpoints ;
  uint   GammaRed;
  uint   GammaGreen;
  uint   GammaBlue;
  uint   Intent;
  uint   ProfileData;
  uint   ProfileSize;
  uint   Reserved;
} ;

enum FMTtype{
  UNKNOWN ,
  CORE ,
  INFO ,
  V4 ,
  V5 ,
} ;

void readvar(T)(File infile, out T x){T[1] b ;T[] d = infile.rawRead(b) ; x=d[0] ; return;}

class FileBMP
{
  File   infile ;
  int    width ;
  int    height ;
  int    bpp ;
  int    curline ;
  BitmapFileHeader bmpfh ;
  BitmapInformationHeader bi ;
  FMTtype  fmttype ;
  uint[]   pallet ;
  uint[][] imgbuf ;

  this(){
  }

  ~this(){
    for( int y=0 ; y<imgbuf.length ; y++ ){
      delete imgbuf[y] ;
    }
    delete imgbuf ;
    delete pallet ;
  }

  int getWidth(){
    return(width) ;
  }

  int getHeight(){
    return(height) ;
  }

  int getBPP(){
    return(bpp) ;
  }

  int open(string fname, string mode="r"){
    infile = File(fname) ;
    int w,h,b ;
    if( mode == "r" ){
      if( readinfo(&w,&h,&b)!=0 ){
        return(-1) ;
      }
      width =w ;
      height=h ;
      bpp   =b ;
      curline=0 ;
    }
    return(0) ;
  }

  int readinfo(int *w, int *h, int *b){
    *w=0 ;
    *h=0 ;
    *b=0 ;
    fmttype = FMTtype.UNKNOWN ;

    infile.readvar(bmpfh.bfType) ;
    infile.readvar(bmpfh.bfSize) ;
    infile.readvar(bmpfh.bfReserved1) ;
    infile.readvar(bmpfh.bfReserved2) ;
    infile.readvar(bmpfh.bfOffBits) ;

    if( ((bmpfh.bfType & 0x00ff)   )== 'B' &&
        ((bmpfh.bfType & 0xff00)>>8)== 'M' ){
      uint infohdsize ;
      infile.readvar(infohdsize) ;
      if( infohdsize==12  || infohdsize==40  ||
          infohdsize==52  || infohdsize==56  || infohdsize==96 || infohdsize==108 ||
          infohdsize==112 || infohdsize==120 || infohdsize==124 ){
        fmttype = FMTtype.CORE ;
        bi.Size=infohdsize ;
        infile.readvar(bi.Width) ;
        infile.readvar(bi.Height) ;
        infile.readvar(bi.Planes) ;
        infile.readvar(bi.BitCount) ;
        *w = bi.Width ;
        *h = bi.Height * (bi.Height<0 ? -1 : 1)  ;
        *b = 24 ;
      }else{
        return(-1) ;
      }

      if( infohdsize==40  ||
          infohdsize==52  || infohdsize==56  || infohdsize==96 || infohdsize==108 ||
          infohdsize==112 || infohdsize==120 || infohdsize==124 ){
        fmttype = FMTtype.INFO ;
        infile.readvar(bi.Compression) ;
        infile.readvar(bi.SizeImage) ;
        infile.readvar(bi.XPixPerMeter) ;
        infile.readvar(bi.YPixPerMeter) ;
        infile.readvar(bi.ClrUsed) ;
        infile.readvar(bi.CirImportant) ;
      }
      if( infohdsize==52  || infohdsize==56  || infohdsize==96 || infohdsize==108 ||
          infohdsize==112 || infohdsize==120 || infohdsize==124 ){
        fmttype = FMTtype.V4 ;
        infile.readvar(bi.RedMask) ;
        infile.readvar(bi.GreenMask) ;
        infile.readvar(bi.BlueMask) ;
      }
      if( infohdsize==56  || infohdsize==96 || infohdsize==108 ||
          infohdsize==112 || infohdsize==120 || infohdsize==124 ){
        infile.readvar(bi.AlphaMask) ;
      }
      if( infohdsize==96 || infohdsize==108 ||
          infohdsize==112 || infohdsize==120 || infohdsize==124 ){
        infile.readvar(bi.CSType) ;
        infile.readvar(bi.Endpoints.ciexyzRed.ciexyzX) ;
        infile.readvar(bi.Endpoints.ciexyzRed.ciexyzY) ;
        infile.readvar(bi.Endpoints.ciexyzRed.ciexyzZ) ;
        infile.readvar(bi.Endpoints.ciexyzGreen.ciexyzX) ;
        infile.readvar(bi.Endpoints.ciexyzGreen.ciexyzY) ;
        infile.readvar(bi.Endpoints.ciexyzGreen.ciexyzZ) ;
        infile.readvar(bi.Endpoints.ciexyzBlue.ciexyzX) ;
        infile.readvar(bi.Endpoints.ciexyzBlue.ciexyzY) ;
        infile.readvar(bi.Endpoints.ciexyzBlue.ciexyzZ) ;
      }
      if( infohdsize==108 ||
          infohdsize==112 || infohdsize==120 || infohdsize==124 ){
        infile.readvar(bi.GammaRed) ;
        infile.readvar(bi.GammaGreen) ;
        infile.readvar(bi.GammaBlue) ;
      }
      if( infohdsize==112 || infohdsize==120 || infohdsize==124 ){
        fmttype = FMTtype.V5 ;
        infile.readvar(bi.Intent) ;
      }
      if( infohdsize==120 || infohdsize==124 ){
        infile.readvar(bi.ProfileData) ;
        infile.readvar(bi.ProfileSize) ;
      }
      if( infohdsize==124 ){
        infile.readvar(bi.Reserved) ;
      }
    }else{
      return(-1) ;
    }

    if( bi.BitCount<=8 ){
      pallet.length= (1<<bi.BitCount) ;
      if( fmttype==FMTtype.CORE ){
        for( int i=0 ; i<pallet.length ; i++ ){
          infile.readvar(pallet[i]) ;
        }
      }else{
        for( int i=0 ; i<pallet.length ; i++ ){
          ubyte R,G,B,X ;
          infile.readvar(B) ;
          infile.readvar(G) ;
          infile.readvar(R) ;
          infile.readvar(X) ;
          pallet[i]=cast(uint)(B)<<16 | cast(uint)(G)<<8 | cast(uint)(R) ;
        }
      }
    }

    if( bi.Compression==BI_RLE8 || bi.Compression==BI_RLE4 ){
      imgbuf = decodeRLE(bi,pallet,infile) ;
    }

    return(0) ;
  }

  uint[][] decodeRLE(BitmapInformationHeader bi, uint[]  pallet, File infile){
    uint[][] imgbuf ;

    imgbuf.length=bi.Height * (bi.Height<0 ? -1 : 1)  ;
    for( int y=0 ; y<imgbuf.length ; y++ ){
      imgbuf[y].length=bi.Width ;
    }

    bool eob=false ;
    for( int y=0 ; y<imgbuf.length ; y++ ){
      for( int x=0 ; ; ){
        ubyte code0, code1 ;
        bool  eol=false ;

        infile.readvar(code0) ;
        infile.readvar(code1) ;
        if( code0==0 ){
          switch(code1){
          case 0  : //EOL
            eol=true ;
            break ;
          case 1  : //EOB
            eob=true ;
            break ;
          case 2  : //move pos
            ubyte offx, offy ;
            infile.readvar(offx) ;
            infile.readvar(offy) ;
            x+=offx ;
            y+=offy ;
            break ;
          default : //absolute mode pix count
            if( bi.Compression==BI_RLE8 ){
              ubyte data ;
              for( int i=0 ; i<code1 ; i++ ){
                infile.readvar(data) ;
                if( x<imgbuf[y].length ){
                  imgbuf[y][x] = pallet[data] ;
                  x++ ;
                }
              }
              if( (code1 & 1)!=0 ){ // skip padding
                infile.readvar(data) ;
              }
            }
            if( bi.Compression==BI_RLE4 ){
              ubyte data ;
              for( int i=0 ; i<code1 ; i++ ){
                infile.readvar(data) ;
                if( x<imgbuf[y].length && i<code1 ){
                  imgbuf[y][x] = pallet[(data&0xf0)>>4] ;
                  x++ ;
                }
                i++ ;
                if( x<imgbuf[y].length && i<code1 ){
                  imgbuf[y][x] = pallet[(data&0x0f)] ;
                  x++ ;
                }
              }
              if( (code1 % 4)==1 || (code1 % 4)==2  ){ // skip padding
                infile.readvar(data) ;
              }
            }
            break ;
          }
          if( eol==true || eob==true ) break ;
        }else{
          if( bi.Compression==BI_RLE8 ){
            for( int i=0 ; i<code0 ; i++ ){
              if( x<imgbuf[y].length ){
                imgbuf[y][x] = pallet[code1] ;
                x++ ;
              }
            }
          }
          if( bi.Compression==BI_RLE4 ){
            for( int i=0 ; i<code0 ; i++ ){
              if( x<imgbuf[y].length ){
                if( x%2==0 ){
                  imgbuf[y][x] = pallet[(code1&0xf0)>>4] ;
                }else{
                  imgbuf[y][x] = pallet[(code1&0x0f)] ;
                }
                x++ ;
              }
            }
          }
        }
      }
      if( eob==true ) break ;
    }

    return(imgbuf) ;
  }

  uint[] readline(int lnum=1){
    uint[]  img ;
    char[]  rbuf ;

    img.length=width*lnum ;

    if( curline<height ){
      if( fmttype!=FMTtype.UNKNOWN ){
        if( bi.Compression==BI_RGB ){
          int recwidth = (((width * bi.BitCount) +31)/32)*4 ;
          for( int y=0 ; y<lnum && curline<height ; y++, curline++ ){
            if( bi.Height>=0 ){
              infile.seek( bmpfh.bfOffBits + (height - curline - 1)*recwidth ) ;
            }
            rbuf = infile.rawRead(new char[recwidth]) ;
            switch( bi.BitCount ){
            case 1 :
              for( int x=0 ; x<width ; x++ ){
                int bytepos = x/8 ;
                int bitpos  = (7 - x%8) ;
                img[width*y+x]=0xff000000 | pallet[(rbuf[bytepos] >> bitpos) & 0x01] ;
              }
              break ;
            case 4 :
              for( int x=0 ; x<width ; x++ ){
                int bytepos = x/2 ;
                int bitpos  = (1-(x%2))*4 ;
                img[width*y+x]=0xff000000 | pallet[(rbuf[bytepos] >> bitpos) & 0x0f] ;
              }
              break ;
            case 8 :
              for( int x=0 ; x<width ; x++ ){
                img[width*y+x]=0xff000000 | pallet[rbuf[x]] ;
              }
              break ;
            case 24 :
              for( int x=0 ; x<width ; x++ ){
                img[width*y+x]=0xff000000 |
                  (rbuf[x*3+0]<<16) |
                  (rbuf[x*3+1]<< 8) |
                  (rbuf[x*3+2]    ) ;
              }
              break ;
            case 32 :
              for( int x=0 ; x<width ; x++ ){
                img[width*y+x]=0xff000000 |
                  (rbuf[x*4+0]<<16) |
                  (rbuf[x*4+1]<< 8) |
                  (rbuf[x*4+2]    ) ;
              }
              break ;
            default :
              break ;
            }
          }
        }

        if( bi.Compression==BI_RLE8 || bi.Compression==BI_RLE4 ){
          for( int y=0 ; y<lnum && curline<height ; y++, curline++ ){
            if( imgbuf[curline].length==0 ){
              writef("%d %d %d\n", imgbuf.length, imgbuf[curline].length, curline) ;
            }
            int sy = bi.Height>=0 ? (height - curline -1) : curline ;
            for( int x=0 ; x<width ; x++ ){
              img[width*y+x]=0xff000000 | imgbuf[sy][x] ;
            }
          }
        }

        if( bi.Compression==BI_BITFIELDS ){
          int recwidth = (((width * bi.BitCount) +31)/32)*4 ;
          for( int y=0 ; y<lnum && curline<height ; y++, curline++ ){
            if( bi.Height>=0 ){
              infile.seek( bmpfh.bfOffBits + (height - curline - 1)*recwidth ) ;
            }
            rbuf = infile.rawRead(new char[recwidth]) ;
            switch( bi.BitCount ){
            case 16 :
              if(       bi.RedMask==0x7c00 && bi.GreenMask==0x03e00 && bi.BlueMask==0x001f ){
                for( int x=0 ; x<width ; x++ ){
                  ushort pix = rbuf[x*2+0] | (rbuf[x*2+1]<<8) ;
                  img[width*y+x]=0xff000000 |
                    ((pix & 0x7c00)>>8) |
                    ((pix & 0x03e0)>>2) |
                    ((pix & 0x001f)<<3) ;
                }
              }else if( bi.RedMask==0xf800 && bi.GreenMask==0x07e00 && bi.BlueMask==0x001f ){
                for( int x=0 ; x<width ; x++ ){
                  ushort pix = rbuf[x*2+0] | (rbuf[x*2+1]<<8) ;
                  img[width*y+x]=0xff000000 |
                    ((pix & 0xf800)>>8) |
                    ((pix & 0x03e0)>>3) |
                    ((pix & 0x001f)<<3) ;
                }
              }
              break ;
            case 32 :
              if( bi.RedMask==0xff000000 && bi.GreenMask==0x00ff0000 && bi.BlueMask==0x0000ff00 ){
                for( int x=0 ; x<width ; x++ ){
                  uint pix = rbuf[x*4+0] | (rbuf[x*4+3]<<8) | (rbuf[x*4+2]<<16) | (rbuf[x*4+1]<<24) ;
                  img[width*y+x]=0xff000000 |
                    ((pix & 0xff000000)>>8) |
                    ((pix & 0x00ff0000)>>8) |
                    ((pix & 0x0000ff00)>>8) ;
                }
              }
              else if( bi.RedMask==0x00ff0000 && bi.GreenMask==0x0000ff00 && bi.BlueMask==0x000000ff ){
                for( int x=0 ; x<width ; x++ ){
                  uint pix = rbuf[x*4+2] | (rbuf[x*4+1]<<8) | (rbuf[x*4+0]<<16) | (rbuf[x*4+3]<<24) ;
                  img[width*y+x]=0xff000000 |
                    (pix & 0x00ff0000) |
                    (pix & 0x0000ff00) |
                    (pix & 0x000000ff) ;
                }
              }
              break ;
            default :
              break ;
            }
          }
        }
      }
    }
    return(img) ;
  }

  int close(){
    infile.detach() ;
    return(0) ;
  }
}

version (BMP_DEBUG){
int main(string[] args)
{
  FileBMP bmp = new FileBMP() ;
  bmp.open(args[1]) ;
  writef("w=%d h=%d b=%d\n", bmp.width, bmp.height, bmp.bpp) ;
  int w = bmp.getWidth() ;
  int h = bmp.getHeight() ;

  File   outfile ;
  char[] cimg ;
  outfile = File("outputtest.ppm","wb") ;
  outfile.writef("P6\n%d %d\n255\n",w,h) ;

  cimg.length=w*h*3 ;
  uint[] l ;
  l = bmp.readline(h) ;
  for( int x=0 ; x<w*h ; x++ ){
    cimg[x*3+2] = (l[x] & 0x00ff0000)>>16 ;
    cimg[x*3+1] = (l[x] & 0x0000ff00)>>8  ;
    cimg[x*3+0] = (l[x] & 0x000000ff)     ;
  }
  outfile.rawWrite(cimg) ;

  outfile.detach() ;

  return(0) ;
}
}
