import std.stdio;
import std.string;
import std.conv;
import std.file;

struct TGAHEADER
{
  ubyte  IDlength ;
  ubyte  colormaptype ;
  ubyte  imagetype ;
  ushort colormaporigin ;
  ushort colormaplength ;
  ubyte  colormapdepth ;
  ushort originX ;
  ushort originY ;
  ushort width ;
  ushort height ;
  ubyte  depth ;
  ubyte  descriptor ;
}

enum {
  NONE         = 0,
  COLORMAP     = 1,
  RGB          = 2,
  GRAYSCALE    = 3,
  RLE_COLORMAP = 9,
  RLE_RGB      =10,
  RLE_GRAYSCALE=11,
}

void readvar(T)(File infile, out T x){T[1] b ;T[] d = infile.rawRead(b) ; x=d[0] ; return;}

class FileTGA
{
  File     infile ;
  TGAHEADER header ;
  int      width ;
  int      height ;
  int      bpp ;
  int      curline ;
  int      headersize=18;
  uint[]   colmap ;
  uint[][] imgbuf ;

  this(){
  }

  ~this(){
    for( int i=0 ; i<imgbuf.length ; i++ ){
      delete imgbuf[i] ;
    }
    delete imgbuf ;
    delete colmap ;
  }

  int getWidth(){
    return(width) ;
  }

  int getHeight(){
    return(height) ;
  }

  int getBPP(){
    return(bpp) ;
  }

  int open(string fname, string mode="r"){
    infile = File(fname) ;
    int w,h,b ;
    if( mode == "r" ){
      if( readinfo(&w,&h,&b)!=0 ){
        return(-1) ;
      }
      width =w ;
      height=h ;
      bpp   =b ;
      curline=0 ;
//      writef("header=%s\n",header) ;
    }
    return(0) ;
  }

  int readinfo(int *w, int *h, int *b){
    infile.readvar(header.IDlength) ;
    infile.readvar(header.colormaptype) ;
    infile.readvar(header.imagetype) ;
    infile.readvar(header.colormaporigin) ;
    infile.readvar(header.colormaplength) ;
    infile.readvar(header.colormapdepth) ;
    infile.readvar(header.originX) ;
    infile.readvar(header.originY) ;
    infile.readvar(header.width) ;
    infile.readvar(header.height) ;
    infile.readvar(header.depth) ;
    infile.readvar(header.descriptor) ;
    *w = header.width ;
    *h = header.height ;
    *b = header.depth ;

    colmap.length = header.colormaplength ;
    for( int i=0 ; i<colmap.length ; i++ ){
      switch( header.colormapdepth ){
      case 24:
        ubyte red,green,blue ;
        infile.readvar(red); infile.readvar(green); infile.readvar(blue);
        colmap[i] = 0xff000000 | red<<16 | green<<8 | blue ;
        break ;
      case 32 :
        ubyte red,green,blue,alpha ;
        infile.readvar(alpha); infile.readvar(red); infile.readvar(green); infile.readvar(blue);
        colmap[i] = alpha<<24 | red<<16 | green<<8 | blue ;
        break ;
      default : break ;
      }
    }

    if( header.imagetype==RLE_COLORMAP ||
        header.imagetype==RLE_RGB      ||
        header.imagetype==RLE_GRAYSCALE ){
      imgbuf = decodeRLE(header,infile) ;
    }
    return(0) ;
  }

  uint[][] decodeRLE(TGAHEADER header, File infile){
      int      ptr=0 ;
      uint[][] imgbuf ;
      imgbuf.length = header.height ;
      for( int i=0 ; i<imgbuf.length ; i++ ){
        imgbuf[i].length=header.width ;
      }

      while( ptr < header.height*header.width ){
        ubyte ctl ; infile.readvar(ctl) ;
        int   count = (ctl & 0x7f) + 1 ;
        if( (ctl & 0x80)!=0 ){
          uint col=readcolor(header.imagetype,header.depth) ;
          for( int i=0 ; i<count ; i++ ){
            imgbuf[ptr/header.width][ptr%header.width]=col ;
            ptr++ ;
          }
        }else{
          for( int i=0 ; i<count ; i++ ){
            imgbuf[ptr/header.width][ptr%header.width]=readcolor(header.imagetype,header.depth) ;
            ptr++ ;
          }
        }
      }
      return(imgbuf) ;
  }

  private uint readcolor(int imagetype, int depth){
    uint col ;

    switch( imagetype ){
    case RLE_COLORMAP :
      ubyte idx ;
      infile.readvar(idx) ;
      col=colmap[idx] ;
      break ;

    case RLE_RGB :
      switch( depth ){
      case 24 :
        ubyte red,green,blue ;
        infile.readvar(red); infile.readvar(green); infile.readvar(blue);
        col = 0xff000000 | red<<16 | green<<8 | blue ;
        break ;
      case 32 :
        ubyte red,green,blue,alpha ;
        infile.readvar(red); infile.readvar(green); infile.readvar(blue); infile.readvar(alpha);
        col = alpha<<24 | red<<16 | green<<8 | blue ;
        break ;
      default :
        break ;
      }
      break ;

    case RLE_GRAYSCALE :
      switch( depth ){
      case 8 :
        ubyte c ;
        infile.readvar(c);
        col = 0xff000000 | c<<16 | c<<8 | c ;
        break ;
      case 16 :
        ubyte c,alpha ;
        infile.readvar(c); infile.readvar(alpha) ;
        col = alpha<<24 | c<<16 | c<<8 | c ;
        break ;
      default :
        break ;
      }
      break ;

    default :
      break ;
    }

    return(col) ;
  }

  uint[] readline(int lnum=1){
    uint[]  img ;
    char[]  rbuf ;
    img.length=width*lnum ;

    if( curline<height ){
      switch( header.imagetype ){
      case RLE_COLORMAP :
      case RLE_RGB :
      case RLE_GRAYSCALE :
        bool dtou= (header.descriptor & 0x20)==0 ;
        bool ltor= (header.descriptor & 0x10)==0 ;
        for( int y=0 ; y<lnum && curline<height ; y++, curline++ ){
          int srcypos ;
          if( dtou==true ){
            srcypos=height - curline - 1 ;
          }else{
            srcypos=curline ;
          }
          for( int x=0 ; x<width ; x++ ){
            img[width*y+x] = imgbuf[srcypos][x] ;
          }
        }
        break ;

      case COLORMAP :
        int  colmapsize = header.colormapdepth/8 * header.colormaplength;
        bool dtou= (header.descriptor & 0x20)==0 ;
        bool ltor= (header.descriptor & 0x10)==0 ;
        for( int y=0 ; y<lnum && curline<height ; y++, curline++ ){
          if( dtou==true ){
            infile.seek( headersize + colmapsize + (height - curline - 1)*width ) ;
          }
          rbuf = infile.rawRead(new char[width]) ;
          for( int x=0 ; x<width ; x++ ){
            img[width*y+x] = colmap[rbuf[x]] ;
          }
        }
        break ;

      case RGB :
        int  Bpp = bpp/8 ;
        bool dtou= (header.descriptor & 0x20)==0 ;
        bool ltor= (header.descriptor & 0x10)==0 ;
        for( int y=0 ; y<lnum && curline<height ; y++, curline++ ){
          if( dtou==true ){
            infile.seek( headersize + (height - curline - 1)*width*Bpp ) ;
          }
          rbuf = infile.rawRead(new char[width*Bpp]) ;
          for( int x=0 ; x<width ; x++ ){
            img[width*y+x] = (bpp==32 ? (rbuf[x*Bpp+3]<<24) : 0xff000000) |
              (rbuf[x*Bpp+0]<<16) |
              (rbuf[x*Bpp+1]<< 8) |
              (rbuf[x*Bpp+2]) ;
          }
        }
        break ;

      case GRAYSCALE :
        int  Bpp = bpp/8 ;
        bool dtou= (header.descriptor & 0x20)==0 ;
        bool ltor= (header.descriptor & 0x10)==0 ;
        for( int y=0 ; y<lnum && curline<height ; y++, curline++ ){
          if( dtou==true ){
            infile.seek( headersize + (height - curline - 1)*width*Bpp ) ;
          }
          rbuf = infile.rawRead(new char[width*Bpp]) ;
          for( int x=0 ; x<width ; x++ ){
            img[width*y+x] = (bpp==16 ? (rbuf[x*Bpp+1]<<24) : 0xff000000) |
              (rbuf[x*Bpp+0]<<16) |
              (rbuf[x*Bpp+0]<< 8) |
              (rbuf[x*Bpp+0]) ;
          }
        }
        break ;

      default :
        break ;
      }
    }

    return(img) ;
  }

  int close(){
    infile.detach() ;
    return(0) ;
  }
}

version (TGA_DEBUG){
int main(string[] args)
{
  FileTGA tga = new FileTGA() ;
  tga.open(args[1]) ;
  writef("w=%d h=%d b=%d\n", tga.width, tga.height, tga.bpp) ;
  int w = tga.getWidth() ;
  int h = tga.getHeight() ;

  File    outfile ;
  char[]  cimg ;
  outfile = File("outputtest.ppm","wb") ;
  outfile.writef("P6\n%d %d\n255\n",w,h) ;
//  cimg.length=w*3 ;
//
//  for( int y=0 ; y<h ; y++ ){
//    uint[] l ;
//    l = tga.readline(1) ;
//    for( int x=0 ; x<w ; x++ ){
//      cimg[x*3+2] = (l[x] & 0x00ff0000)>>16 ;
//      cimg[x*3+1] = (l[x] & 0x0000ff00)>>8  ;
//      cimg[x*3+0] = (l[x] & 0x000000ff)     ;
//    }
//    outfile.writeString(cimg) ;
//  }

  cimg.length=w*h*3 ;
  uint[] l ;
  l = tga.readline(h) ;
  for( int x=0 ; x<w*h ; x++ ){
    cimg[x*3+2] = (l[x] & 0x00ff0000)>>16 ;
    cimg[x*3+1] = (l[x] & 0x0000ff00)>>8  ;
    cimg[x*3+0] = (l[x] & 0x000000ff)     ;
  }
  outfile.rawWrite(cimg) ;

  outfile.detach() ;

  return(0) ;
}
}
