import std.stdio;
import std.string;
import std.path;
import std.utf;
import util.FileTGA;
import util.FileBMP;
import util.FileDDS;
import util.FilePNM;

struct JPG_HANDLE{
  uint *a ;
  uint *b ;
  uint *c ;
  uint *d ;
}

struct PNG_HANDLE{
  uint *a ;
  uint *b ;
  uint *c ;
  uint *d ;
  int   A ;
  int   B ;
  uint *e ;
  int   C ;
  int   D ;
}

enum FileOPstat{
  IDLE,
  JPGRDFOPEN,
  PNGRDFOPEN,
  PNGRDMOPEN,
  TGARDFOPEN,
  BMPRDFOPEN,
  DDSRDFOPEN,
  PNMRDFOPEN,
  JPGWTFOPEN,
  PNGWTFOPEN
}


extern (C) int readinfo_png( wchar *file, int *width, int *height, int *bpp, ubyte *mempng, int memlength ) ;
extern (C) PNG_HANDLE *load_png_preproc( wchar *file, ubyte *mempng, int memlength ) ;
extern (C) int load_png_getline( uint *img, PNG_HANDLE *png_handle ) ;
extern (C) int load_png_postproc( PNG_HANDLE *png_handle ) ;

extern (C) PNG_HANDLE *save_png_preproc( wchar *file, int width, int height ) ;
extern (C) int save_png_putline( uint *img, PNG_HANDLE *png_handle ) ;
extern (C) int save_png_postproc( PNG_HANDLE *png_handle ) ;

extern (C) int readinfo_jpg( wchar *file, int *width, int *height, int *bpp );
extern (C) JPG_HANDLE *load_jpg_preproc( wchar *file ) ;
extern (C) int load_jpg_getline( uint *img, JPG_HANDLE *jpg_handle ) ;
extern (C) int load_jpg_postproc( JPG_HANDLE *jpg_handle ) ;

extern (C) void imagelib_error_handler(){
  throw new Exception("Error in ImageLibrary") ;
  return ;
}

class Image {
  int         width  ;
  int         height ;
  int         bpp    ;
  int         curline;
  JPG_HANDLE *jpg_handle ;
  PNG_HANDLE *png_handle ;
  FileOPstat  state  ;
  FileTGA     tga ;
  FileBMP     bmp ;
  FileDDS     dds ;
  FilePNM     pnm ;
  string jpgext  = ".jpg" ;
  string jpgext2 = ".jpeg" ;
  string pngext  = ".png" ;
  string tgaext  = ".tga" ;
  string bmpext  = ".bmp" ;
  string ddsext  = ".dds" ;
  string ppmext  = ".ppm" ;
  string pgmext  = ".pgm" ;
  string pbmext  = ".pbm" ;

  this(){
    width  = 0 ;
    height = 0 ;
    bpp    = 0 ;
    state  = FileOPstat.IDLE ;
    jpg_handle = null ;
    png_handle = null ;
    tga    = null ;
    bmp    = null ;
    dds    = null ;
    pnm    = null ;
  }

  ~this(){
    close() ;
  }

  int getWidth(){
    return(width) ;
  }

  int getHeight(){
    return(height) ;
  }

  int getBPP(){
    return(bpp) ;
  }

  int setWidth(int width){
    if( state == FileOPstat.IDLE ){
      this.width=width ;
    }
    return(0) ;
  }

  int setHeight(int height){
    if( state == FileOPstat.IDLE ){
      this.height=height ;
    }
    return(0) ;
  }

  int setBPP(int bpp){
    if( state == FileOPstat.IDLE ){
      this.bpp=bpp ;
    }
    return(0) ;
  }

  bool[string] getKnownExtlist(){
    bool[string] extlist ;
    extlist[jpgext] =true ;
    extlist[jpgext2]=true ;
    extlist[pngext] =true ;
    extlist[tgaext] =true ;
    extlist[bmpext] =true ;
    extlist[ddsext] =true ;
    extlist[ppmext] =true ;
    extlist[pgmext] =true ;
    extlist[pbmext] =true ;
    return(extlist) ;
  }

  int open( string fname, string mode="r", string type="" ){
    int    w,h,b ;

    if( state!=FileOPstat.IDLE ){
      close() ;
    }

    string ext=toLower(extension(fname)) ;

    if( type=="jpg" || (type=="" && (ext==jpgext || ext==jpgext2)) ){
      if( mode == "r" ){
        if( readinfo_jpg( toUTFz!(wchar*)(fname), &w, &h, &b)!=0 ){
          return(-1) ;
        }
        jpg_handle = load_jpg_preproc(toUTFz!(wchar*)(fname)) ;
        if( jpg_handle==null ) return(-1) ;
        state  = FileOPstat.JPGRDFOPEN ;
        width  = w ;
        height = h ;
        bpp    = b ;
        curline= 0 ;
      }
      return(0) ;
    }

    if( type=="png" || (type=="" && (ext==pngext)) ){
      if( mode == "r" ){
        if( readinfo_png(toUTFz!(wchar*)(fname), &w, &h, &b, null, 0)!=0 ){
          return(-1) ;
        }
        png_handle = load_png_preproc(toUTFz!(wchar*)(fname), null, 0) ;
        if( png_handle==null ) return(-1) ;
        state  = FileOPstat.PNGRDFOPEN ;
        width  = w ;
        height = h ;
        bpp    = b ;
        curline= 0 ;
      }else{
        if( width<=0 || height<=0 ) return(-1) ;
        png_handle = save_png_preproc(toUTFz!(wchar*)(fname), width, height) ;
        if( png_handle==null ) return(-1) ;
        state  = FileOPstat.PNGWTFOPEN ;
        curline= 0 ;
      }
      return(0) ;
    }

    if( type=="tga" || (type=="" && (ext==tgaext)) ){
      if( mode == "r" ){
        tga = new FileTGA() ;
        if( tga.open(fname)<0 ) return(-1) ;
        state  = FileOPstat.TGARDFOPEN ;
        width  = tga.getWidth() ;
        height = tga.getHeight() ;
        bpp    = tga.getBPP() ;
        curline= 0 ;
      }
      return(0) ;
    }

    if( type=="bmp" || (type=="" && (ext==bmpext)) ){
      if( mode == "r" ){
        bmp = new FileBMP() ;
        if( bmp.open(fname)<0 ) return(-1) ;
        state  = FileOPstat.BMPRDFOPEN ;
        width  = bmp.getWidth() ;
        height = bmp.getHeight() ;
        bpp    = bmp.getBPP() ;
        curline= 0 ;
      }
      return(0) ;
    }

    if( type=="dds" || (type=="" && (ext==ddsext)) ){
      if( mode == "r" ){
        dds = new FileDDS() ;
        if( dds.open(fname)<0 ) return(-1) ;
        state  = FileOPstat.DDSRDFOPEN ;
        width  = dds.getWidth() ;
        height = dds.getHeight() ;
        bpp    = dds.getBPP() ;
        curline= 0 ;
      }
      return(0) ;
    }

    if( type=="ppm" || (type=="" && (ext==ppmext)) ){
      if( mode == "r" ){
        pnm = new FilePNM() ;
        if( pnm.open(fname)<0 ) return(-1) ;
        state  = FileOPstat.PNMRDFOPEN ;
        width  = pnm.getWidth() ;
        height = pnm.getHeight() ;
        bpp    = pnm.getBPP() ;
        curline= 0 ;
      }
      return(0) ;
    }

    if( type=="pgm" || (type=="" && (ext==pgmext)) ){
      if( mode == "r" ){
        pnm = new FilePNM() ;
        if( pnm.open(fname)<0 ) return(-1) ;
        state  = FileOPstat.PNMRDFOPEN ;
        width  = pnm.getWidth() ;
        height = pnm.getHeight() ;
        bpp    = pnm.getBPP() ;
        curline= 0 ;
      }
      return(0) ;
    }

    if( type=="pbm" || (type=="" && (ext==pbmext)) ){
      if( mode == "r" ){
        pnm = new FilePNM() ;
        if( pnm.open(fname)<0 ) return(-1) ;
        state  = FileOPstat.PNMRDFOPEN ;
        width  = pnm.getWidth() ;
        height = pnm.getHeight() ;
        bpp    = pnm.getBPP() ;
        curline= 0 ;
      }
      return(0) ;
    }

    return(-1) ;
  }

  int open( ubyte[] memdt, string type, string mode="r" ){
    int    w,h,b ;

    if( mode == "r" ){
      switch(type){
      case "png" :
        if( readinfo_png(null, &w, &h, &b, memdt.ptr, cast(int)(memdt.length))!=0 ){
          return(-1) ;
        }
        png_handle = load_png_preproc( null, memdt.ptr, cast(int)(memdt.length) ) ;
        if( png_handle==null ) return(-1) ;
        state  = FileOPstat.PNGRDMOPEN ;
        width  = w ;
        height = h ;
        bpp    = b ;
        curline= 0 ;
        return(0) ;
        break ;
      default :
        break ;
      }
    }
    return(-1) ;
  }

  uint[] readline(int lnum=1, bool reverse_order=false){
    uint[] img ;

    switch( state ){
    case FileOPstat.JPGRDFOPEN :
      img.length=width*lnum ;
      for( int i=0 ; i<lnum ; i++ ){
        load_jpg_getline( &img[i*width], jpg_handle) ;
        curline++ ;
      }
      break ;
    case FileOPstat.PNGRDFOPEN :
    case FileOPstat.PNGRDMOPEN :
      img.length=width*lnum ;
      for( int i=0 ; i<lnum ; i++ ){
        load_png_getline( &img[i*width], png_handle) ;
        curline++ ;
      }
      break ;
    case FileOPstat.TGARDFOPEN :
      img = tga.readline(lnum) ;
      curline+=lnum ;
      break ;
    case FileOPstat.BMPRDFOPEN :
      img = bmp.readline(lnum) ;
      curline+=lnum ;
      break ;
    case FileOPstat.DDSRDFOPEN :
      img = dds.readline(lnum) ;
      curline+=lnum ;
      break ;
    case FileOPstat.PNMRDFOPEN :
      img = pnm.readline(lnum) ;
      curline+=lnum ;
      break ;
    default :
      break ;
    }

    if( reverse_order==true ){
      foreach( ref pix ; img ){
        pix = (pix&0xff00ff00) |  ((pix&0x00ff0000)>>16) |  ((pix&0x000000ff)<<16) ;
      }
    }

    return(img) ;
  }

  int writeline(uint[] img, int lnum=1, bool reverse_order=false){
    if( reverse_order==true ){
      foreach( ref pix ; img ){
        pix = (pix&0xff00ff00) |  ((pix&0x00ff0000)>>16) |  ((pix&0x000000ff)<<16) ;
      }
    }
    switch( state ){
    case FileOPstat.PNGWTFOPEN :
      for( int i=0 ; i<lnum ; i++ ){
        save_png_putline( &img[i*width], png_handle) ;
        curline++ ;
      }
      break ;
    default :
      return(-1) ;
      break ;
    }

    return(0) ;
  }

  int close(){
    switch( state ){
    case FileOPstat.JPGRDFOPEN :
      load_jpg_postproc( jpg_handle ) ;
      jpg_handle=null ;
      break ;
    case FileOPstat.PNGRDFOPEN :
    case FileOPstat.PNGRDMOPEN :
//      if( curline<height ){
//      uint[] trashbuf = new uint[width] ;
//      for( int i=curline ; i<height ; i++ ){
//        load_png_getline( trashbuf.ptr, png_handle) ;
//      }
//      }
      load_png_postproc( png_handle ) ;
      png_handle=null ;
      break ;
    case FileOPstat.TGARDFOPEN :
      tga.close() ;
      break ;
    case FileOPstat.BMPRDFOPEN :
      bmp.close() ;
      break ;
    case FileOPstat.DDSRDFOPEN :
      dds.close() ;
      break ;
    case FileOPstat.PNMRDFOPEN :
      pnm.close() ;
      break ;
    case FileOPstat.PNGWTFOPEN :
      save_png_postproc( png_handle ) ;
      png_handle=null ;
      break ;
    default :
      break ;
    }
    state     =FileOPstat.IDLE ;
    width     = 0 ;
    height    = 0 ;
    bpp       = 0 ;
    curline   = 0 ;

    return(0) ;
  }

}

version (CLASS_DEBUG){

int main(string[] args)
{
  Image img = new Image() ;

  if( args.length!=2 ){
    writef("Usage : %s filename\n",args[0]) ;
    return(0) ;
  }

  if( img.open(args[1])==0 ){
    File     outfile ;
    uint[]   iimg ;
    char[]   cimg ;

    outfile = File("test.ppm","wb") ;
    outfile.writef("P6\n%d %d\n255\n",img.getWidth(),img.getHeight()) ;

    cimg.length=img.getWidth()*3 ;
    for( int i=0 ; i<img.getHeight() ; i++ ){
      iimg = img.readline() ;
      for( int j=0 ; j<iimg.length ; j++ ){
        cimg[j*3+2] = (iimg[j] & 0x00ff0000)>>16 ;
        cimg[j*3+1] = (iimg[j] & 0x0000ff00)>>8  ;
        cimg[j*3+0] = (iimg[j] & 0x000000ff)     ;
      }
      outfile.rawWrite(cimg) ;
    }

    outfile.detach() ;

  }else{
    writef("file open error (%s)\n",args[1]) ;
  }
  return(0) ;
}
}
