/**********************************************************************
  PNGロード
  $Id:$
**********************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "png.h"

struct PNG_HANDLE{
  png_structp    pp     ; /* PNG read pointer  */
  png_infop      info   ; /* PNG info pointers */
  FILE          *fp     ;
  unsigned char *lnbuf  ;
  int            width  ;
  int            bpp    ;
  unsigned char *mtop   ;
  int            mcur   ;
  int            isfile ;
} ;

void png_memrdproc( png_structp pp, png_bytep dst, png_size_t length)
{
  struct PNG_HANDLE *ph = (struct PNG_HANDLE*)png_get_io_ptr(pp) ;
  memcpy(dst, &ph->mtop[ph->mcur], length) ;
  ph->mcur += length ;
}

int readinfo_png( wchar_t *file, int *width, int *height, int *bpp, unsigned char *mempng, int memlength )
{
  FILE          *fp=NULL ;  /* File pointer */
  png_structp   pp;   /* PNG read pointer */
  png_infop     info; /* PNG info pointers */
  struct PNG_HANDLE png_handle ;
  int           bit_depth ;
  int           color_type ;
  int           interlace_type ;
  int           ret=0 ;

  *width =0 ;
  *height=0 ;
  *bpp   =0 ;

  if( file==NULL && mempng==NULL ) return(-1) ;

  png_handle.isfile=0 ;
  if( file!=NULL ){
    if( (fp=_wfopen(file, L"rb"))==NULL ){
      //    printf("pngload image file open error\n") ;
      return(-1) ;
    }
    png_handle.isfile=1 ;
  }

  pp   = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if( pp!=NULL ){
    info = png_create_info_struct(pp);
    if( info!=NULL ){
      if( png_handle.isfile==1 ){
        png_init_io( pp, fp );
      }else{
        png_handle.mtop=mempng ;
        png_handle.mcur=0 ;
        png_set_read_fn(pp, &png_handle, png_memrdproc) ;
      }
      png_read_info( pp, info );
      png_get_IHDR(pp, info, (png_uint_32*)width, (png_uint_32*)height, &bit_depth, &color_type, &interlace_type, NULL, NULL);
      png_destroy_read_struct(&pp, &info, NULL);
      *bpp  = (color_type & PNG_COLOR_MASK_COLOR)!=0 ? bit_depth*3 : 0 ;
      *bpp += (color_type & PNG_COLOR_MASK_ALPHA)!=0 ? 8 : 0 ;
    }else{
      ret=-1 ;
      //printf("png_create_info_struct faild\n") ;
      png_destroy_read_struct(&pp, NULL, NULL);
    }
  }else{
    ret=-1 ;
    //printf("png_create_read_struct faild\n") ;
  }

  if( png_handle.isfile==1 ) fclose(fp);

  return(ret);
}


struct PNG_HANDLE *load_png_preproc( wchar_t *file, unsigned char *mempng, int memlength  )
{
  struct PNG_HANDLE *png_handle ;
  png_uint_32 width , height ;
  int         bit_depth ;
  int         color_type ;
  int         interlace_type ;

  if( file==NULL && mempng==NULL ) return(NULL) ;

  if( (png_handle = malloc(sizeof(struct PNG_HANDLE)))==NULL ){
    return(NULL) ;
  }

  png_handle->isfile=0 ;
  if( file !=NULL ){
    if( (png_handle->fp=_wfopen(file, L"rb"))==NULL ){
      free(png_handle) ;
      return(NULL) ;
    }
    png_handle->isfile=1 ;
  }

  png_handle->pp   = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  png_handle->info = png_create_info_struct(png_handle->pp);

  if( png_handle->isfile==1 ){
    png_init_io( png_handle->pp, png_handle->fp );
  }else{
    png_handle->mtop=mempng ;
    png_handle->mcur=0 ;
    png_set_read_fn(png_handle->pp, png_handle, png_memrdproc) ;
  }

  png_read_info( png_handle->pp, png_handle->info );
  png_get_IHDR(png_handle->pp, png_handle->info, &width, &height, &bit_depth, &color_type, &interlace_type, NULL, NULL);

  if( bit_depth < 8 ){
    png_set_packing(png_handle->pp);
    png_set_expand(png_handle->pp);
    if( png_get_valid(png_handle->pp, png_handle->info, PNG_INFO_sBIT) ){
      png_color_8p sig_bit_p;
      png_get_sBIT(png_handle->pp, png_handle->info, &sig_bit_p);
      png_set_shift(png_handle->pp, sig_bit_p);
    }
  }else{
    if( bit_depth == 16 ) png_set_strip_16(png_handle->pp);
  }

  switch( color_type ){
  case PNG_COLOR_TYPE_PALETTE :
    png_set_palette_to_rgb(png_handle->pp);
    break ;
  case PNG_COLOR_TYPE_GRAY :
    png_set_gray_to_rgb(png_handle->pp);
    break ;
  default :
    break ;
  }

  png_handle->bpp  = 24 ;
  png_handle->bpp += (color_type & PNG_COLOR_MASK_ALPHA)!=0 ? 8 : 0 ;
  png_handle->lnbuf = malloc( width * png_handle->bpp/8 ) ;
  if( png_handle->lnbuf==NULL ){
    png_destroy_read_struct(&png_handle->pp, &png_handle->info, NULL);
    if( png_handle->isfile==1 ) fclose( png_handle->fp ) ;
    free(   png_handle     ) ;
    return(NULL) ;
  }
  png_handle->width=width ;
  return(png_handle) ;
}


int load_png_getline( unsigned int  *img, struct PNG_HANDLE *png_handle )
{
  int i ;
  png_bytepp   *pixel_buf[1] ;
  png_bytepp  **pixels ;

  pixels = &pixel_buf[0] ;
  switch( png_handle->bpp ){
  case 24 :
    pixel_buf[0]=(png_bytepp *)&png_handle->lnbuf[0] ;
    png_read_rows( png_handle->pp, (png_bytepp)pixels, NULL, 1 );
    for( i=0 ; i<png_handle->width ; i++ ){
      img[i]=0xff000000 |
             (png_handle->lnbuf[i*3+2])<<16 |
             (png_handle->lnbuf[i*3+1])<<8  |
             (png_handle->lnbuf[i*3+0]) ;
    }
    break ;
  case 32 :
    pixel_buf[0]=(png_bytepp *)&png_handle->lnbuf[0] ;
    png_read_rows( png_handle->pp, (png_bytepp)pixels, NULL, 1 );
    for( i=0 ; i<png_handle->width ; i++ ){
      img[i]=(png_handle->lnbuf[i*4+3])<<24 |
             (png_handle->lnbuf[i*4+2])<<16 |
             (png_handle->lnbuf[i*4+1])<<8  |
             (png_handle->lnbuf[i*4+0]) ;
    }
    break ;
  default :
    break ;
  }
  return(0) ;
}


int load_png_postproc( struct PNG_HANDLE *png_handle )
{
  //png_read_end(png_handle->pp, NULL) ;
  png_destroy_read_struct(&png_handle->pp, &png_handle->info, NULL) ;
  if( png_handle->isfile==1 ) fclose( png_handle->fp ) ;
  free(  png_handle->lnbuf) ;
  free(  png_handle       ) ;
  png_handle=NULL ;

  return(0) ;
}


#ifdef DEBUG
int main(int argc , char *argv[])
{
  char *image ;
  int  width,height,bpp ;
  FILE *fp ;

  if( argc!=2 ){
    printf("usage : %s file\n",argv[0]) ;
    exit(0) ;
  }
  printf("libpng version = %d\n",PNG_LIBPNG_VER) ;

  printf("--------- test1 ---------\n") ;
  if( readinfo_png(argv[1], &width, &height, &bpp)!=0 ){
    printf("PNG file open error\n") ;
  }else{
    printf("%s w=%d h=%d bpp=%d\n",argv[1],width,height,bpp) ;
  }
  struct PNG_HANDLE *png_handle ;
  png_handle = load_png_preproc(argv[1]) ;
  if( png_handle==NULL ){
    printf("PNG file open error\n") ;
    exit(0) ;
  }
  int i,j ;
  unsigned int  *iimg ;
  unsigned char *cimg ;
  iimg=malloc(sizeof(unsigned int)*width) ;
  cimg=malloc(width*3) ;
  fp=fopen("test2.ppm","wb") ;
  fputs( "P6" , fp ) ; fputc( 0x0a , fp ) ;
  fprintf( fp , "%d %d", width, height) ;
  fputc( 0x0a , fp ) ;
  fputs( "255" , fp ) ; fputc( 0x0a , fp ) ;

  for( i=0 ; i<height ; i++ ){
    load_png_getline( iimg, png_handle ) ;
    for( j=0 ; j<width ; j++ ){
      cimg[j*3+0] = (iimg[j] & 0x000000ff)     ;
      cimg[j*3+1] = (iimg[j] & 0x0000ff00)>>8  ;
      cimg[j*3+2] = (iimg[j] & 0x00ff0000)>>16 ;
    }
    fwrite( cimg , width*3 , 1 , fp ) ;
  }
  fclose(fp) ;
  load_png_postproc( png_handle ) ;

  exit(0) ;
}
#endif
