/**********************************************************************
  PNGセーブ
  $Id:$
**********************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include "png.h"

struct PNG_HANDLE{
  png_structp    pp    ; /* PNG read pointer  */
  png_infop      info  ; /* PNG info pointers */
  FILE          *fp    ;
  unsigned char *lnbuf ;
  int            width ;
  int            bpp   ;
} ;


struct PNG_HANDLE *save_png_preproc( wchar_t *file, int width, int height )
{
  struct PNG_HANDLE *png_handle ;

  if( (png_handle = malloc(sizeof(struct PNG_HANDLE)))==NULL ){
    return(NULL) ;
  }

  if( (png_handle->fp=_wfopen(file, L"wb"))==NULL ){
    free(png_handle) ;
    return(NULL) ;
  }

  png_handle->pp   = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  png_handle->info = png_create_info_struct(png_handle->pp);

  png_handle->bpp  = 24 ;
  png_handle->lnbuf = malloc( width * png_handle->bpp/8 ) ;
  if( png_handle->lnbuf==NULL ){
    png_destroy_write_struct(&png_handle->pp, &png_handle->info);
    fclose( png_handle->fp ) ;
    free(   png_handle     ) ;
    return(NULL) ;
  }
  png_handle->width=width ;

  png_init_io( png_handle->pp, png_handle->fp );
  png_set_compression_level(png_handle->pp, 9);
  png_set_IHDR(png_handle->pp, png_handle->info,
               width, height, 8, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
               PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT) ;

  png_set_sRGB(  png_handle->pp, png_handle->info, PNG_sRGB_INTENT_SATURATION);
  png_write_info( png_handle->pp, png_handle->info ) ;

  return(png_handle) ;
}


int save_png_putline(  unsigned int  *img, struct PNG_HANDLE *png_handle )
{
  int i ;
  png_bytepp   *pixel_buf[1] ;
  png_bytepp  **pixels ;

  pixels = &pixel_buf[0] ;
  switch( png_handle->bpp ){
  case 24 :
    for( i=0 ; i<png_handle->width ; i++ ){
      png_handle->lnbuf[i*3+0] = (img[i]>>16) & 0xff ;
      png_handle->lnbuf[i*3+1] = (img[i]>> 8) & 0xff ;
      png_handle->lnbuf[i*3+2] = (img[i]    ) & 0xff ;
    }
    pixel_buf[0]=(png_bytepp *)&png_handle->lnbuf[0] ;
    png_write_rows( png_handle->pp, (png_bytepp)pixels, 1 );
    break ;
  default :
    break ;
  }
  return(0) ;
}


int save_png_postproc( struct PNG_HANDLE *png_handle )
{
  png_write_end(png_handle->pp, png_handle->info) ;
  png_destroy_write_struct(&png_handle->pp, NULL) ;
  fclose(png_handle->fp   ) ;
  free(  png_handle->lnbuf) ;
  free(  png_handle       ) ;
  png_handle=NULL ;

  return(0) ;
}


#ifdef DEBUG
int main(int argc , char *argv[])
{
  char *image ;
  int  width,height,bpp ;
  FILE *fp ;

  if( argc!=2 ){
    printf("usage : %s file\n",argv[0]) ;
    exit(0) ;
  }
  printf("libpng version = %d\n",PNG_LIBPNG_VER) ;

  printf("--------- test1 ---------\n") ;
  struct PNG_HANDLE *png_handle ;
  width =640 ;
  height=480 ;
  png_handle = save_png_preproc(argv[1], width, height) ;
  if( png_handle==NULL ){
    printf("PNG write file open error\n") ;
    exit(0) ;
  }
  int i,j ;
  unsigned int  *wimg ;
  wimg=malloc(sizeof(unsigned int)*width) ;

  for( i=0 ; i<height ; i++ ){
    for( j=0 ; j<width ; j++ ){
      wimg[j] = 0x00ff0000 ;
    }
    save_png_putline( wimg, png_handle ) ;
  }
  save_png_postproc( png_handle ) ;
  free(wimg) ;

  exit(0) ;
}
#endif
