import win32.windows;
import std.string;
import std.stdio;
import std.utf;
import win32frame.writek;
import win32frame.WindowObject ;

class Split : WindowObject {
  int maxobjnum    ;
  int split_mode   ;
  int sash_width=5 ;
  int mouseX    =0 ;
  int mouseY    =0 ;
  int mouseB    =0 ;
  int split_pos    ;
  int curpos       ;
  int cappos       ;
  int touching     ;
  int minpos       ;
  int maxpos       ;
  bool onMouseDrag ;

  this(string name, int mode=0) {
    winc.style          = 0;
    this.setStyle(WS_CHILD | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE) ;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.lpszMenuName   = null;
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    winc.hbrBackground  = CreateSolidBrush(GetSysColor(COLOR_3DFACE)) ;
    winc.lpszClassName  = toUTFz!(wchar*)("Split") ;
    super(name) ;
    this.registwinc() ;
    maxobjnum =2 ;
    split_mode=mode ;
    touching=false ;
    split_pos=-1  ;
    minpos=0     ;
    maxpos=65535 ;
    onMouseDrag=false ;
  }

  override int addObject(WindowObject WObj, HMENU hmenu, bool fitWidth=false, bool fitHeight=false ){
    int  ret=-1 ;
    if( WOBJ_List.length<=maxobjnum ){
      ret = WObj.create(this.getHWND(), hmenu) ;
      if( WObj.getHWND() != null ){
        WOBJ_List.length = WOBJ_List.length + 1 ;
        WOBJ_List[(WOBJ_List.length-1)] = WObj ;
        sendmsg( WM_SIZE, 0, (winH<<16 | winW) );
      }
    }
    return( ret ) ;
  }

  int setSashPos(int pos, int min_pos, int max_pos){
    if( min_pos<=max_pos && min_pos<=pos || max_pos>=pos){
      if( min_pos>=0 ){
        minpos=min_pos ;
        maxpos=max_pos ;
        split_pos=pos ;
      }
    }
    return(0) ;
  }

  int setCurPos(int newpos){
    if( newpos>=minpos && newpos<=maxpos ){
      split_pos=newpos ;
      curpos   =newpos ;
    }
    sendmsg( WM_SIZE, 0, (winH<<16 | winW) );
    return(0) ;
  }

  int getCurPos(){
    return(curpos) ;
  }

  int setOnMouseDrag(bool enable=false){
    onMouseDrag=enable ;
    return(0) ;
  }

  int reDraw(){
    sendmsg(WM_SIZE, 0, (winH<<16 | winW) );
    return(0) ;
  }

  override LRESULT preproc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch( msg ){
    case WM_CREATE :
      if( split_pos<0 ){
        split_pos = split_mode==0 ? winW/2 : winH/2 ;
      }
      curpos = split_pos ;
      break ;

    case WM_MOVE :
      _winX=LOWORD(lp) ;
      _winY=HIWORD(lp) ;
      break ;

    case WM_LBUTTONDOWN:
    case WM_LBUTTONUP:
    case WM_MOUSEMOVE:
      POINT point;

      GetCursorPos(&point);
      ScreenToClient( hwnd, &point );
      mouseX = point.x ;
      mouseY = point.y ;
      mouseX = mouseX<0 ? 0 : mouseX>=winW ? winW-1 : mouseX ;
      mouseY = mouseY<0 ? 0 : mouseY>=winH ? winH-1 : mouseY ;
      if( msg==WM_LBUTTONDOWN || msg==WM_MOUSEMOVE ){ //マウスカーソル形状を変更する
        if( minpos!=split_pos || maxpos!=split_pos ){
          if( split_mode==0 ){
            SetCursor( LoadCursor(null, cast(wchar*)(IDC_SIZEWE) ) ) ;
          }else{
            SetCursor( LoadCursor(null, cast(wchar*)(IDC_SIZENS) ) ) ;
          }
        }
      }
      if( msg==WM_LBUTTONDOWN ){
        SetCapture( hwnd ) ;
        touching=true ;
        cappos = split_mode==0 ? mouseX : mouseY ;
      }
      if( msg==WM_MOUSEMOVE ){
        if( touching==true ){
          curpos = split_pos + (split_mode==0 ? mouseX : mouseY ) - cappos ;
          curpos = curpos < minpos ? minpos : curpos>maxpos ? maxpos : curpos ;
          if( split_mode==0 ){
            if( (winW - sash_width)<curpos ) curpos=winW - sash_width ;
          }else{
            if( (winH - sash_width)<curpos ) curpos=winH - sash_width ;
          }
          sendmsg( WM_SIZE, wp, (winH<<16 | winW) );
          if( onMouseDrag==true ){
            sendmsgtoparent( WM_COMMAND, cast(WPARAM)(this.getHMENU()), 0) ;
          }
        }
      }
      if( msg==WM_LBUTTONUP   ){
        ReleaseCapture() ;
        split_pos = curpos ;
        touching=false ;
        if( onMouseDrag==true ){
          sendmsgtoparent( WM_COMMAND, cast(WPARAM)(this.getHMENU()), 0) ;
        }
      }
      break ;

    case WM_SIZE :
      _winW=LOWORD(lp) ;
      _winH=HIWORD(lp) ;
      if( split_mode==0 ){
        int primary_window_width = (curpos) ;
        int primary_window_height= (winH) ;
        int pos_x = 0 ;
        int pos_y = 0 ;

        foreach( i, WOBJ ; WOBJ_List ){
          if( WOBJ !is null ){
            if( i==0 ){
              MoveWindow( WOBJ.getHWND(), pos_x, pos_y,
                          primary_window_width, primary_window_height, false ) ;
            }else{
              MoveWindow( WOBJ.getHWND(), pos_x, pos_y,
                          (winW - sash_width ) - primary_window_width, primary_window_height, false ) ;
            }
            pos_x += primary_window_width + sash_width ;
          }
        }
        RedrawWindow( getParentHWND(), null, null, RDW_INVALIDATE|RDW_ALLCHILDREN ) ;
      }else{
        int primary_window_width = (winW) ;
        int primary_window_height= (curpos) ;
        int pos_x = 0 ;
        int pos_y = 0 ;

        foreach( i, WOBJ ; WOBJ_List ){
          if( WOBJ !is null ){
            if( i==0 ){
              MoveWindow( WOBJ.getHWND(), pos_x, pos_y,
                          primary_window_width, primary_window_height, false ) ;
            }else{
              MoveWindow( WOBJ.getHWND(), pos_x, pos_y,
                          primary_window_width, (winH - sash_width) - primary_window_height, false ) ;
            }
            pos_y += primary_window_height + sash_width ;
          }
        }
        RedrawWindow( getParentHWND(), null, null, RDW_INVALIDATE|RDW_ALLCHILDREN ) ;
      }
      break ;

    case WM_NOTIFY:
    case WM_HSCROLL:
    case WM_VSCROLL:
    case WM_COMMAND:
      sendmsgtoparent(msg, wp,lp) ;
      break ;

    default : break ;
    }

    return( this.proc(hwnd , msg , wp , lp) ) ;
  }

}
